#ifndef ETHSERP_PARSER
#define ETHSERP_PARSER

#include <stdio.h>
#include <iostream>
#include <vector>
#include <map>
#include "util.h"

// Serpent text -> parse tree
Node parseSerpent(std::string s);

#endif
